package sun.security.provider;

import java.security.PublicKey;
import sun.security.util.KeyUtil;
import jdk.internal.event.EventHelper;
import sun.security.util.Pem;
import sun.security.pkcs.ParsingException;
import java.util.Arrays;
import sun.security.pkcs.PKCS7;
import java.util.ArrayList;
import java.io.PushbackInputStream;
import java.security.cert.CRL;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import sun.security.provider.certpath.X509CertPath;
import java.io.ByteArrayInputStream;
import java.security.cert.CertPath;
import java.security.cert.CRLException;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.io.ByteArrayOutputStream;
import sun.security.provider.certpath.X509CertificatePair;
import java.security.cert.CertificateException;
import java.io.IOException;
import java.security.cert.Certificate;
import java.io.InputStream;
import sun.security.x509.X509CRLImpl;
import sun.security.x509.X509CertImpl;
import sun.security.util.Cache;
import java.security.cert.CertificateFactorySpi;

public class X509Factory extends CertificateFactorySpi
{
    public static final String BEGIN_CERT = "-----BEGIN CERTIFICATE-----";
    public static final String END_CERT = "-----END CERTIFICATE-----";
    private static final int ENC_MAX_LENGTH = 4194304;
    private static final Cache<Object, X509CertImpl> certCache;
    private static final Cache<Object, X509CRLImpl> crlCache;
    
    public X509Factory() {
    }
    
    @Override
    public Certificate engineGenerateCertificate(InputStream inputStream) throws CertificateException {
        if (inputStream == null) {
            X509Factory.certCache.clear();
            X509CertificatePair.clearCache();
            throw new CertificateException("Missing input stream");
        }
        try {
            byte[] oneBlock = readOneBlock(inputStream);
            if (oneBlock != null) {
                X509CertImpl x509CertImpl = (X509CertImpl)getFromCache(X509Factory.certCache, oneBlock);
                if (x509CertImpl != null) {
                    return x509CertImpl;
                }
                X509CertImpl x509CertImpl2 = new X509CertImpl(oneBlock);
                addToCache(X509Factory.certCache, x509CertImpl2.getEncodedInternal(), x509CertImpl2);
                this.recordEvent(x509CertImpl2);
                return x509CertImpl2;
            }
            else {
                throw new IOException("Empty input");
            }
        }
        catch (IOException cause) {
            throw new CertificateException("Could not parse certificate: " + cause.toString(), cause);
        }
    }
    
    private static int readFully(InputStream inputStream, ByteArrayOutputStream byteArrayOutputStream, int i) throws IOException {
        int n = 0;
        byte[] array = new byte[2048];
        while (i > 0) {
            int read = inputStream.read(array, 0, (i < 2048) ? i : 2048);
            if (read <= 0) {
                break;
            }
            byteArrayOutputStream.write(array, 0, read);
            n += read;
            i -= read;
        }
        return n;
    }
    
    public static synchronized X509CertImpl intern(X509Certificate x509Certificate) throws CertificateException {
        if (x509Certificate == null) {
            return null;
        }
        int n = x509Certificate instanceof X509CertImpl;
        byte[] certData;
        if (!(n == 0)) {
            certData = ((X509CertImpl)x509Certificate).getEncodedInternal();
        }
        else {
            certData = x509Certificate.getEncoded();
        }
        X509CertImpl x509CertImpl = (X509CertImpl)getFromCache(X509Factory.certCache, certData);
        if (x509CertImpl != null) {
            return x509CertImpl;
        }
        X509CertImpl x509CertImpl2;
        if (!(n == 0)) {
            x509CertImpl2 = (X509Certificate)x509Certificate;
        }
        else {
            x509CertImpl2 = new X509CertImpl(certData);
            certData = x509CertImpl2.getEncodedInternal();
        }
        addToCache(X509Factory.certCache, certData, x509CertImpl2);
        return x509CertImpl2;
    }
    
    public static synchronized X509CRLImpl intern(X509CRL x509CRL) throws CRLException {
        if (x509CRL == null) {
            return null;
        }
        int n = x509CRL instanceof X509CRLImpl;
        byte[] crlData;
        if (!(n == 0)) {
            crlData = ((X509CRLImpl)x509CRL).getEncodedInternal();
        }
        else {
            crlData = x509CRL.getEncoded();
        }
        X509CRLImpl x509CRLImpl = (X509CRLImpl)getFromCache(X509Factory.crlCache, crlData);
        if (x509CRLImpl != null) {
            return x509CRLImpl;
        }
        X509CRLImpl x509CRLImpl2;
        if (!(n == 0)) {
            x509CRLImpl2 = (X509CRL)x509CRL;
        }
        else {
            x509CRLImpl2 = new X509CRLImpl(crlData);
            crlData = x509CRLImpl2.getEncodedInternal();
        }
        addToCache(X509Factory.crlCache, crlData, x509CRLImpl2);
        return x509CRLImpl2;
    }
    
    private static synchronized <K, V> V getFromCache(Cache<K, V> cache, byte[] b) {
        return cache.get(new Cache.EqualByteArray(b));
    }
    
    private static synchronized <V> void addToCache(Cache<Object, V> cache, byte[] b, V v) {
        if (b.length > 4194304) {
            return;
        }
        cache.put(new Cache.EqualByteArray(b), v);
        return;
    }
    
    @Override
    public CertPath engineGenerateCertPath(InputStream inputStream) throws CertificateException {
        if (inputStream == null) {
            throw new CertificateException("Missing input stream");
        }
        try {
            byte[] oneBlock = readOneBlock(inputStream);
            if (oneBlock != null) {
                return new X509CertPath(new ByteArrayInputStream(oneBlock));
            }
            throw new IOException("Empty input");
        }
        catch (IOException ex) {
            throw new CertificateException(ex.getMessage());
        }
    }
    
    @Override
    public CertPath engineGenerateCertPath(InputStream inputStream, String encoding) throws CertificateException {
        if (inputStream == null) {
            throw new CertificateException("Missing input stream");
        }
        try {
            byte[] oneBlock = readOneBlock(inputStream);
            if (oneBlock != null) {
                return new X509CertPath(new ByteArrayInputStream(oneBlock), encoding);
            }
            throw new IOException("Empty input");
        }
        catch (IOException ex) {
            throw new CertificateException(ex.getMessage());
        }
    }
    
    @Override
    public CertPath engineGenerateCertPath(List<? extends Certificate> certs) throws CertificateException {
        return new X509CertPath(certs);
    }
    
    @Override
    public Iterator<String> engineGetCertPathEncodings() {
        return X509CertPath.getEncodingsStatic();
    }
    
    @Override
    public Collection<? extends Certificate> engineGenerateCertificates(InputStream inputStream) throws CertificateException {
        if (inputStream == null) {
            throw new CertificateException("Missing input stream");
        }
        try {
            return this.parseX509orPKCS7Cert(inputStream);
        }
        catch (IOException cause) {
            throw new CertificateException(cause);
        }
    }
    
    @Override
    public CRL engineGenerateCRL(InputStream inputStream) throws CRLException {
        if (inputStream == null) {
            X509Factory.crlCache.clear();
            throw new CRLException("Missing input stream");
        }
        try {
            byte[] oneBlock = readOneBlock(inputStream);
            if (oneBlock != null) {
                X509CRLImpl x509CRLImpl = (X509CRLImpl)getFromCache(X509Factory.crlCache, oneBlock);
                if (x509CRLImpl != null) {
                    return x509CRLImpl;
                }
                X509CRLImpl x509CRLImpl2 = new X509CRLImpl(oneBlock);
                addToCache(X509Factory.crlCache, x509CRLImpl2.getEncodedInternal(), x509CRLImpl2);
                return x509CRLImpl2;
            }
            else {
                throw new IOException("Empty input");
            }
        }
        catch (IOException ex) {
            throw new CRLException(ex.getMessage());
        }
    }
    
    @Override
    public Collection<? extends CRL> engineGenerateCRLs(InputStream inputStream) throws CRLException {
        if (inputStream == null) {
            throw new CRLException("Missing input stream");
        }
        try {
            return this.parseX509orPKCS7CRL(inputStream);
        }
        catch (IOException ex) {
            throw new CRLException(ex.getMessage());
        }
    }
    
    private Collection<? extends Certificate> parseX509orPKCS7Cert(InputStream in) throws CertificateException, IOException {
        PushbackInputStream pushbackInputStream = new PushbackInputStream(in);
        ArrayList<X509CertImpl> list = new ArrayList();
        int read = pushbackInputStream.read();
        if (read == -1) {
            return new ArrayList(0);
        }
        pushbackInputStream.unread(read);
        byte[] array = readOneBlock(pushbackInputStream);
        if (array == null) {
            throw new CertificateException("No certificate data found");
        }
        try {
            X509Certificate[] certificates = new PKCS7(array).getCertificates();
            if (certificates != null) {
                return Arrays.asList(certificates);
            }
            return new ArrayList(0);
        }
        catch (ParsingException ex) {
            while (array != null) {
                list.add(new X509CertImpl(array));
                array = readOneBlock(pushbackInputStream);
            }
            return list;
        }
    }
    
    private Collection<? extends CRL> parseX509orPKCS7CRL(InputStream in) throws CRLException, IOException {
        PushbackInputStream pushbackInputStream = new PushbackInputStream(in);
        ArrayList<X509CRLImpl> list = new ArrayList();
        int read = pushbackInputStream.read();
        if (read == -1) {
            return new ArrayList(0);
        }
        pushbackInputStream.unread(read);
        byte[] array = readOneBlock(pushbackInputStream);
        if (array == null) {
            throw new CRLException("No CRL data found");
        }
        try {
            X509CRL[] crLs = new PKCS7(array).getCRLs();
            if (crLs != null) {
                return Arrays.asList(crLs);
            }
            return new ArrayList(0);
        }
        catch (ParsingException ex) {
            while (array != null) {
                list.add(new X509CRLImpl(array));
                array = readOneBlock(pushbackInputStream);
            }
            return list;
        }
    }
    
    private static byte[] readOneBlock(InputStream inputStream) throws IOException {
        int read = inputStream.read();
        if (read == -1) {
            return null;
        }
        if (read == 48) {
            ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream(2048);
            byteArrayOutputStream.write(read);
            readBERInternal(inputStream, byteArrayOutputStream, read);
            return byteArrayOutputStream.toByteArray();
        }
        char[] copy = new char[2048];
        int count = 0;
        int n = (read == 45) ? 1 : 0;
        read = ((read == 45) ? -1 : read);
        while (true) {
            int read2 = inputStream.read();
            if (read2 == -1) {
                return null;
            }
            if (read2 == 45) {
                ++n;
            }
            else {
                n = 0;
                read = read2;
            }
            if (n == 5 && (read == -1 || read == 13 || read == 10)) {
                StringBuilder o = new StringBuilder("-----");
                int n2;
                while (true) {
                    int read3 = inputStream.read();
                    if (read3 == -1) {
                        throw new IOException("Incomplete data");
                    }
                    if (read3 == 10) {
                        n2 = 10;
                        break;
                    }
                    if (read3 == 13) {
                        int read4 = inputStream.read();
                        if (read4 == -1) {
                            throw new IOException("Incomplete data");
                        }
                        if (read4 == 10) {
                            n2 = 10;
                            break;
                        }
                        n2 = 13;
                        copy[count++] = (char)read4;
                        break;
                    }
                    else {
                        o.append((char)read3);
                    }
                }
                while (true) {
                    int read5 = inputStream.read();
                    if (read5 == -1) {
                        throw new IOException("Incomplete data");
                    }
                    if (read5 != 45) {
                        copy[count++] = (char)read5;
                        if (count >= copy.length) {
                            copy = Arrays.copyOf(copy, copy.length + 1024);
                        }
                        else {
                            continue;
                        }
                    }
                    else {
                        StringBuilder o2 = new StringBuilder("-");
                        while (true) {
                            int read6 = inputStream.read();
                            if (read6 != -1 && read6 != n2 && read6 != 10) {
                                if (read6 != 13) {
                                    o2.append((char)read6);
                                }
                                else {
                                    continue;
                                }
                            }
                            else {
                                break;
                            }
                        }
                        checkHeaderFooter(o.toString(), o2.toString());
                        return Pem.decode(new String(copy, 0, count));
                    }
                }
            }
            else {
                continue;
            }
        }
    }
    
    private static void checkHeaderFooter(String s, String s2) throws IOException {
        if (s.length() < 16 || !s.startsWith("-----BEGIN ") || !s.endsWith("-----")) {
            throw new IOException("Illegal header: " + s);
        }
        if (s2.length() < 14 || !s2.startsWith("-----END ") || !s2.endsWith("-----")) {
            throw new IOException("Illegal footer: " + s2);
        }
        if (!s.substring(11, s.length() - 5).equals(s2.substring(9, s2.length() - 5))) {
            throw new IOException("Header and footer do not match: " + s + " " + s2);
        }
        return;
    }
    
    private static int readBERInternal(InputStream inputStream, ByteArrayOutputStream byteArrayOutputStream, int read) throws IOException {
        if (read == -1) {
            read = inputStream.read();
            if (read == -1) {
                throw new IOException("BER/DER tag info absent");
            }
            if ((read & 0x1F) == 0x1F) {
                throw new IOException("Multi octets tag not supported");
            }
            byteArrayOutputStream.write(read);
        }
        int n = inputStream.read();
        if (n == -1) {
            throw new IOException("BER/DER length info absent");
        }
        byteArrayOutputStream.write(n);
        if (n == 128) {
            if ((read & 0x20) != 0x20) {
                throw new IOException("Non constructed encoding must have definite length");
            }
            while (readBERInternal(inputStream, byteArrayOutputStream, -1) != 0) {}
        }
        else {
            if (n < 128) {
            }
            else if (n == 129) {
                n = inputStream.read();
                if (n == -1) {
                    throw new IOException("Incomplete BER/DER length info");
                }
                byteArrayOutputStream.write(n);
            }
            else if (n == 130) {
                int read9 = inputStream.read();
                int read10 = inputStream.read();
                if (read10 == -1) {
                    throw new IOException("Incomplete BER/DER length info");
                }
                byteArrayOutputStream.write(read9);
                byteArrayOutputStream.write(read10);
                n = (read9 << 8 | read10);
            }
            else if (n == 131) {
                int read6 = inputStream.read();
                int read7 = inputStream.read();
                int read8 = inputStream.read();
                if (read8 == -1) {
                    throw new IOException("Incomplete BER/DER length info");
                }
                byteArrayOutputStream.write(read6);
                byteArrayOutputStream.write(read7);
                byteArrayOutputStream.write(read8);
                n = (read6 << 16 | read7 << 8 | read8);
            }
            else if (n == 132) {
                int read2 = inputStream.read();
                int read3 = inputStream.read();
                int read4 = inputStream.read();
                int read5 = inputStream.read();
                if (read5 == -1) {
                    throw new IOException("Incomplete BER/DER length info");
                }
                if (read2 > 127) {
                    throw new IOException("Invalid BER/DER data (a little huge?)");
                }
                byteArrayOutputStream.write(read2);
                byteArrayOutputStream.write(read3);
                byteArrayOutputStream.write(read4);
                byteArrayOutputStream.write(read5);
                n = (read2 << 24 | read3 << 16 | read4 << 8 | read5);
            }
            else {
                throw new IOException("Invalid BER/DER data (too huge?)");
            }
            if (readFully(inputStream, byteArrayOutputStream, n) != n) {
                throw new IOException("Incomplete BER/DER data");
            }
        }
        return read;
    }
    
    private void recordEvent(X509CertImpl x509CertImpl) {
        if (EventHelper.isLoggingSecurity()) {
            PublicKey publicKey = x509CertImpl.getPublicKey();
            EventHelper.logX509CertificateEvent(x509CertImpl.getSigAlgName(), x509CertImpl.getSerialNumber().toString(16), x509CertImpl.getSubjectDN().getName(), x509CertImpl.getIssuerDN().getName(), publicKey.getAlgorithm(), KeyUtil.getKeySize(publicKey), (long)x509CertImpl.hashCode(), x509CertImpl.getNotBefore().getTime(), x509CertImpl.getNotAfter().getTime());
        }
    }
    
    static {
        certCache = Cache.newSoftMemoryCache(750);
        crlCache = Cache.newSoftMemoryCache(750);
    }
}
